@@ -63,16 +63,10 @@ QBluetoothDeviceDiscoveryAgentPrivate::QBluetoothDeviceDiscoveryAgentPrivate(
     m_adapterAddress(deviceAdapter),
     pendingCancel(false),
     pendingStart(false),
-    manager(0),
-    adapter(0),
-    managerBluez5(0),
-    adapterBluez5(0),
-    discoveryTimer(0),
     useExtendedDiscovery(false),
     lowEnergySearchTimeout(-1), // remains -1 on BlueZ 4 -> timeout not supported
     q_ptr(parent)
 {
-    Q_Q(QBluetoothDeviceDiscoveryAgent);
     if (isBluez5()) {
         lowEnergySearchTimeout = 20000;
         managerBluez5 = new OrgFreedesktopDBusObjectManagerInterface(
@@ -80,8 +74,11 @@ QBluetoothDeviceDiscoveryAgentPrivate::QBluetoothDeviceDiscoveryAgentPrivate(
                                            QStringLiteral("/"),
                                            QDBusConnection::systemBus(), parent);
         QObject::connect(managerBluez5,
-                         SIGNAL(InterfacesAdded(QDBusObjectPath,InterfaceList)),
-                         q, SLOT(_q_InterfacesAdded(QDBusObjectPath,InterfaceList)));
+                         &OrgFreedesktopDBusObjectManagerInterface::InterfacesAdded,
+                         q_ptr,
+                         [this](const QDBusObjectPath &objectPath, InterfaceList interfacesAndProperties) {
+            this->_q_InterfacesAdded(objectPath, interfacesAndProperties);
+        });
 
         // start private address monitoring
         BluetoothManagement::instance();
@@ -89,8 +86,9 @@ QBluetoothDeviceDiscoveryAgentPrivate::QBluetoothDeviceDiscoveryAgentPrivate(
         manager = new OrgBluezManagerInterface(QStringLiteral("org.bluez"), QStringLiteral("/"),
                                            QDBusConnection::systemBus(), parent);
         QObject::connect(&extendedDiscoveryTimer,
-                         SIGNAL(timeout()),
-                         q, SLOT(_q_extendedDeviceDiscoveryTimeout()));
+                         &QTimer::timeout, q_ptr, [this]() {
+            this->_q_extendedDeviceDiscoveryTimeout();
+        });
         extendedDiscoveryTimer.setInterval(10000);
         extendedDiscoveryTimer.setSingleShot(true);
     }
@@ -137,6 +135,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::start(QBluetoothDeviceDiscoveryAgent
     }
 
     discoveredDevices.clear();
+    devicesProperties.clear();
 
     if (managerBluez5) {
         startBluez5(methods);
@@ -164,22 +163,26 @@ void QBluetoothDeviceDiscoveryAgentPrivate::start(QBluetoothDeviceDiscoveryAgent
                                            QDBusConnection::systemBus());
 
     Q_Q(QBluetoothDeviceDiscoveryAgent);
-    QObject::connect(adapter, SIGNAL(DeviceFound(QString, QVariantMap)),
-                     q, SLOT(_q_deviceFound(QString, QVariantMap)));
-    QObject::connect(adapter, SIGNAL(PropertyChanged(QString, QDBusVariant)),
-                     q, SLOT(_q_propertyChanged(QString, QDBusVariant)));
+    QObject::connect(adapter, &OrgBluezAdapterInterface::DeviceFound,
+                     q, [this](const QString &address, const QVariantMap &dict) {
+        this->_q_deviceFound(address, dict);
+    });
+    QObject::connect(adapter, &OrgBluezAdapterInterface::PropertyChanged,
+                     q, [this](const QString &name, const QDBusVariant &value) {
+        this->_q_propertyChanged(name, value);
+    });
 
     QDBusPendingReply<QVariantMap> propertiesReply = adapter->GetProperties();
     propertiesReply.waitForFinished();
     if (propertiesReply.isError()) {
         errorString = propertiesReply.error().message();
         delete adapter;
-        adapter = 0;
+        adapter = nullptr;
         qCDebug(QT_BT_BLUEZ) << Q_FUNC_INFO << "ERROR: " << errorString;
         lastError = QBluetoothDeviceDiscoveryAgent::InputOutputError;
         Q_Q(QBluetoothDeviceDiscoveryAgent);
         delete adapter;
-        adapter = 0;
+        adapter = nullptr;
         emit q->error(lastError);
         return;
     }
@@ -189,7 +192,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::start(QBluetoothDeviceDiscoveryAgent
         lastError = QBluetoothDeviceDiscoveryAgent::PoweredOffError;
         errorString = QBluetoothDeviceDiscoveryAgent::tr("Device is powered off");
         delete adapter;
-        adapter = 0;
+        adapter = nullptr;
         emit q->error(lastError);
         return;
     }
@@ -218,7 +221,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::start(QBluetoothDeviceDiscoveryAgent
     discoveryReply.waitForFinished();
     if (discoveryReply.isError()) {
         delete adapter;
-        adapter = 0;
+        adapter = nullptr;
         errorString = discoveryReply.error().message();
         lastError = QBluetoothDeviceDiscoveryAgent::InputOutputError;
         Q_Q(QBluetoothDeviceDiscoveryAgent);
@@ -251,7 +254,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::startBluez5(QBluetoothDeviceDiscover
         lastError = QBluetoothDeviceDiscoveryAgent::PoweredOffError;
         errorString = QBluetoothDeviceDiscoveryAgent::tr("Device is powered off");
         delete adapterBluez5;
-        adapterBluez5 = 0;
+        adapterBluez5 = nullptr;
         emit q->error(lastError);
         return;
     }
@@ -276,7 +279,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::startBluez5(QBluetoothDeviceDiscover
             errorString = QBluetoothDeviceDiscoveryAgent::tr("One or more device discovery methods "
                                                              "are not supported on this platform");
             delete adapterBluez5;
-            adapterBluez5 = 0;
+            adapterBluez5 = nullptr;
             emit q->error(lastError);
             return;
         } else if (filterReply.error().type() != QDBusError::UnknownMethod) {
@@ -285,8 +288,10 @@ void QBluetoothDeviceDiscoveryAgentPrivate::startBluez5(QBluetoothDeviceDiscover
     }
 
     QtBluezDiscoveryManager::instance()->registerDiscoveryInterest(adapterBluez5->path());
-    QObject::connect(QtBluezDiscoveryManager::instance(), SIGNAL(discoveryInterrupted(QString)),
-            q, SLOT(_q_discoveryInterrupted(QString)));
+    QObject::connect(QtBluezDiscoveryManager::instance(), &QtBluezDiscoveryManager::discoveryInterrupted,
+                     q, [this](const QString &path){
+        this->_q_discoveryInterrupted(path);
+    });
 
     // collect initial set of information
     QDBusPendingReply<ManagedObjectList> reply = managerBluez5->GetManagedObjects();
@@ -305,7 +310,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::startBluez5(QBluetoothDeviceDiscover
                     if (path.path().indexOf(adapterBluez5->path()) != 0)
                         continue; //devices whose path doesn't start with same path we skip
 
-                    deviceFoundBluez5(path.path());
+                    deviceFoundBluez5(path.path(), jt.value());
                     if (!isActive()) // Can happen if stop() was called from a slot in user code.
                       return;
                 }
@@ -317,8 +322,10 @@ void QBluetoothDeviceDiscoveryAgentPrivate::startBluez5(QBluetoothDeviceDiscover
     if (!discoveryTimer) {
         discoveryTimer = new QTimer(q);
         discoveryTimer->setSingleShot(true);
-        QObject::connect(discoveryTimer, SIGNAL(timeout()),
-                         q, SLOT(_q_discoveryFinished()));
+        QObject::connect(discoveryTimer, &QTimer::timeout,
+                         q, [this]() {
+            this->_q_discoveryFinished();
+        });
     }
 
     if (lowEnergySearchTimeout > 0) { // otherwise no timeout and stop() required
@@ -359,10 +366,12 @@ void QBluetoothDeviceDiscoveryAgentPrivate::_q_deviceFound(const QString &addres
     QBluetoothDeviceInfo device(btAddress, btName, btClass);
     if (dict.value(QStringLiteral("RSSI")).isValid())
         device.setRssi(dict.value(QStringLiteral("RSSI")).toInt());
-    QList<QBluetoothUuid> uuids;
-    foreach (const QString &u, dict.value(QLatin1String("UUIDs")).toStringList())
+    QVector<QBluetoothUuid> uuids;
+    const QStringList uuidStrings
+            = dict.value(QLatin1String("UUIDs")).toStringList();
+    for (const QString &u : uuidStrings)
         uuids.append(QBluetoothUuid(u));
-    device.setServiceUuids(uuids, QBluetoothDeviceInfo::DataIncomplete);
+    device.setServiceUuids(uuids);
     device.setCached(dict.value(QStringLiteral("Cached")).toBool());
 
 
@@ -395,45 +404,23 @@ void QBluetoothDeviceDiscoveryAgentPrivate::_q_deviceFound(const QString &addres
     emit q->deviceDiscovered(device);
 }
 
-void QBluetoothDeviceDiscoveryAgentPrivate::deviceFoundBluez5(const QString& devicePath)
+// Returns invalid QBluetoothDeviceInfo in case of error
+static QBluetoothDeviceInfo createDeviceInfoFromBluez5Device(const QVariantMap& properties)
 {
-    Q_Q(QBluetoothDeviceDiscoveryAgent);
+    const QBluetoothAddress btAddress(properties[QStringLiteral("Address")].toString());
+    if (btAddress.isNull())
+        return QBluetoothDeviceInfo();
 
-    if (!q->isActive())
-        return;
+    const QString btName = properties[QStringLiteral("Alias")].toString();
+    quint32 btClass = properties[QStringLiteral("Class")].toUInt();
 
-    OrgBluezDevice1Interface device(QStringLiteral("org.bluez"), devicePath,
-                                    QDBusConnection::systemBus());
-
-    if (device.adapter().path() != adapterBluez5->path())
-        return;
-
-    const QBluetoothAddress btAddress(device.address());
-    if (btAddress.isNull()) // no point reporting an empty address
-        return;
-
-    const QString btName = device.alias();
-    quint32 btClass = device.classProperty();
-
-    qCDebug(QT_BT_BLUEZ) << "Discovered: " << btAddress.toString() << btName
-                         << "Num UUIDs" << device.uUIDs().count()
-                         << "total device" << discoveredDevices.count() << "cached"
-                         << "RSSI" << device.rSSI() << "Class" << btClass;
-
-    OrgFreedesktopDBusPropertiesInterface *prop = new OrgFreedesktopDBusPropertiesInterface(
-                QStringLiteral("org.bluez"), devicePath, QDBusConnection::systemBus(), q);
-    QObject::connect(prop, SIGNAL(PropertiesChanged(QString,QVariantMap,QStringList)),
-                     q, SLOT(_q_PropertiesChanged(QString,QVariantMap,QStringList)));
-    // remember what we have to cleanup
-    propertyMonitors.append(prop);
-
-    // read information
     QBluetoothDeviceInfo deviceInfo(btAddress, btName, btClass);
-    deviceInfo.setRssi(device.rSSI());
+    deviceInfo.setRssi(qvariant_cast<short>(properties[QStringLiteral("RSSI")]));
 
-    QList<QBluetoothUuid> uuids;
+    QVector<QBluetoothUuid> uuids;
     bool foundLikelyLowEnergyUuid = false;
-    for (const auto &u: device.uUIDs()) {
+    const QStringList foundUuids = qvariant_cast<QStringList>(properties[QStringLiteral("UUIDs")]);
+    for (const auto &u: foundUuids) {
         const QBluetoothUuid id(u);
         if (id.isNull())
             continue;
@@ -447,7 +434,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::deviceFoundBluez5(const QString& dev
         }
         uuids.append(id);
     }
-    deviceInfo.setServiceUuids(uuids, QBluetoothDeviceInfo::DataIncomplete);
+    deviceInfo.setServiceUuids(uuids);
 
     if (!btClass) {
         deviceInfo.setCoreConfigurations(QBluetoothDeviceInfo::LowEnergyCoreConfiguration);
@@ -457,10 +444,56 @@ void QBluetoothDeviceDiscoveryAgentPrivate::deviceFoundBluez5(const QString& dev
             deviceInfo.setCoreConfigurations(QBluetoothDeviceInfo::BaseRateAndLowEnergyCoreConfiguration);
     }
 
+    const ManufacturerDataList deviceManufacturerData = qdbus_cast<ManufacturerDataList>(properties[QStringLiteral("ManufacturerData")]);
+    const QList<quint16> keys = deviceManufacturerData.keys();
+    for (quint16 key : keys)
+        deviceInfo.setManufacturerData(
+                    key, deviceManufacturerData.value(key).variant().toByteArray());
+
+    return deviceInfo;
+}
+
+void QBluetoothDeviceDiscoveryAgentPrivate::deviceFoundBluez5(const QString &devicePath,
+                                                              const QVariantMap &properties)
+{
+    Q_Q(QBluetoothDeviceDiscoveryAgent);
+
+    if (!q->isActive())
+        return;
+
+    auto deviceAdapter = qvariant_cast<QDBusObjectPath>(properties[QStringLiteral("Adapter")]);
+     if (deviceAdapter.path() != adapterBluez5->path())
+         return;
+
+    // read information
+    QBluetoothDeviceInfo deviceInfo = createDeviceInfoFromBluez5Device(properties);
+    if (!deviceInfo.isValid()) // no point reporting an empty address
+        return;
+
+    qCDebug(QT_BT_BLUEZ) << "Discovered: " << deviceInfo.name() << deviceInfo.address()
+                         << "Num UUIDs" << deviceInfo.serviceUuids().count()
+                         << "total device" << discoveredDevices.count() << "cached"
+                         << "RSSI" << deviceInfo.rssi()
+                         << "Num ManufacturerData" << deviceInfo.manufacturerData().size();
+
+    OrgFreedesktopDBusPropertiesInterface *prop = new OrgFreedesktopDBusPropertiesInterface(
+                QStringLiteral("org.bluez"), devicePath, QDBusConnection::systemBus(), q);
+    QObject::connect(prop, &OrgFreedesktopDBusPropertiesInterface::PropertiesChanged,
+                     q, [this](const QString &interface, const QVariantMap &changedProperties,
+                            const QStringList &invalidatedProperties) {
+        this->_q_PropertiesChanged(interface, changedProperties, invalidatedProperties);
+    });
+
+    // remember what we have to cleanup
+    propertyMonitors.append(prop);
+
+    // Cache the properties so we do not have to access dbus every time to get a value
+    devicesProperties[devicePath] = properties;
+
     for (int i = 0; i < discoveredDevices.size(); i++) {
         if (discoveredDevices[i].address() == deviceInfo.address()) {
-            if (discoveredDevices[i] == deviceInfo && lowEnergySearchTimeout > 0) {
-                qCDebug(QT_BT_BLUEZ) << "Duplicate: " << btAddress.toString();
+            if (lowEnergySearchTimeout > 0 && discoveredDevices[i] == deviceInfo) {
+                qCDebug(QT_BT_BLUEZ) << "Duplicate: " << deviceInfo.address();
                 return;
             }
             discoveredDevices.replace(i, deviceInfo);
@@ -484,13 +517,13 @@ void QBluetoothDeviceDiscoveryAgentPrivate::_q_propertyChanged(const QString &na
             Q_Q(QBluetoothDeviceDiscoveryAgent);
             if (pendingCancel && !pendingStart) {
                 adapter->deleteLater();
-                adapter = 0;
+                adapter = nullptr;
 
                 pendingCancel = false;
                 emit q->canceled();
             } else if (pendingStart) {
                 adapter->deleteLater();
-                adapter = 0;
+                adapter = nullptr;
 
                 pendingStart = false;
                 pendingCancel = false;
@@ -514,7 +547,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::_q_propertyChanged(const QString &na
                 QDBusPendingReply<> reply = adapter->StopDiscovery();
                 reply.waitForFinished();
                 adapter->deleteLater();
-                adapter = 0;
+                adapter = nullptr;
                 emit q->finished();
             }
         } else {
@@ -529,7 +562,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::_q_extendedDeviceDiscoveryTimeout()
 
     if (adapter) {
         adapter->deleteLater();
-        adapter = 0;
+        adapter = nullptr;
     }
     if (isActive()) {
         Q_Q(QBluetoothDeviceDiscoveryAgent);
@@ -548,7 +581,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::_q_InterfacesAdded(const QDBusObject
     if (interfaces_and_properties.contains(QStringLiteral("org.bluez.Device1"))) {
         // device interfaces belonging to different adapter
         // will be filtered out by deviceFoundBluez5();
-        deviceFoundBluez5(object_path.path());
+        deviceFoundBluez5(object_path.path(), interfaces_and_properties[QStringLiteral("org.bluez.Device1")]);
     }
 }
 
@@ -566,7 +599,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::_q_discoveryFinished()
     propertyMonitors.clear();
 
     delete adapterBluez5;
-    adapterBluez5 = 0;
+    adapterBluez5 = nullptr;
 
     if (pendingCancel && !pendingStart) {
         pendingCancel = false;
@@ -599,7 +632,7 @@ void QBluetoothDeviceDiscoveryAgentPrivate::_q_discoveryInterrupted(const QStrin
         // does this automatically when emitting discoveryInterrupted(QString) signal
 
         delete adapterBluez5;
-        adapterBluez5 = 0;
+        adapterBluez5 = nullptr;
 
         errorString = QBluetoothDeviceDiscoveryAgent::tr("Bluetooth adapter error");
         lastError = QBluetoothDeviceDiscoveryAgent::InputOutputError;
@@ -609,28 +642,89 @@ void QBluetoothDeviceDiscoveryAgentPrivate::_q_discoveryInterrupted(const QStrin
 
 void QBluetoothDeviceDiscoveryAgentPrivate::_q_PropertiesChanged(const QString &interface,
                                                                  const QVariantMap &changed_properties,
-                                                                 const QStringList &)
+                                                                 const QStringList &invalidated_properties)
 {
     Q_Q(QBluetoothDeviceDiscoveryAgent);
-    if (interface == QStringLiteral("org.bluez.Device1")
-            && changed_properties.contains(QStringLiteral("RSSI"))) {
-        OrgFreedesktopDBusPropertiesInterface *props =
-                qobject_cast<OrgFreedesktopDBusPropertiesInterface *>(q->sender());
-        if (!props)
-            return;
+    if (interface != QStringLiteral("org.bluez.Device1"))
+        return;
+
+    OrgFreedesktopDBusPropertiesInterface *props =
+            qobject_cast<OrgFreedesktopDBusPropertiesInterface *>(q->sender());
+    if (!props)
+        return;
+
+    const QString path = props->path();
+    if (!devicesProperties.contains(path))
+        return;
+
+    // Update the cached properties before checking changed_properties for RSSI and ManufacturerData
+    // so the cached properties are always up to date.
+    QVariantMap & properties = devicesProperties[path];
+    for (QVariantMap::const_iterator it = changed_properties.constBegin();
+         it != changed_properties.constEnd(); ++it) {
+        properties[it.key()] = it.value();
+    }
+
+    for (const QString & property : invalidated_properties)
+        properties.remove(property);
+
+    const auto info = createDeviceInfoFromBluez5Device(properties);
+    if (!info.isValid())
+        return;
+
+    if (changed_properties.contains(QStringLiteral("RSSI"))
+        || changed_properties.contains(QStringLiteral("ManufacturerData"))) {
 
-        OrgBluezDevice1Interface device(QStringLiteral("org.bluez"), props->path(),
-                                            QDBusConnection::systemBus());
         for (int i = 0; i < discoveredDevices.size(); i++) {
-            if (discoveredDevices[i].address().toString() == device.address()) {
-                qCDebug(QT_BT_BLUEZ) << "Updating RSSI for" << device.address()
-                         << changed_properties.value(QStringLiteral("RSSI"));
-                discoveredDevices[i].setRssi(
-                            changed_properties.value(QStringLiteral("RSSI")).toInt());
+            if (discoveredDevices[i].address() == info.address()) {
+                QBluetoothDeviceInfo::Fields updatedFields = QBluetoothDeviceInfo::Field::None;
+                if (changed_properties.contains(QStringLiteral("RSSI"))) {
+                    qCDebug(QT_BT_BLUEZ) << "Updating RSSI for" << info.address()
+                                         << changed_properties.value(QStringLiteral("RSSI"));
+                    discoveredDevices[i].setRssi(
+                                changed_properties.value(QStringLiteral("RSSI")).toInt());
+                    updatedFields.setFlag(QBluetoothDeviceInfo::Field::RSSI);
+                }
+                if (changed_properties.contains(QStringLiteral("ManufacturerData"))) {
+                    qCDebug(QT_BT_BLUEZ) << "Updating ManufacturerData for" << info.address();
+                    ManufacturerDataList changedManufacturerData =
+                            qdbus_cast< ManufacturerDataList >(changed_properties.value(QStringLiteral("ManufacturerData")));
+
+                    const QList<quint16> keys = changedManufacturerData.keys();
+                    bool wasNewValue = false;
+                    for (quint16 key : keys) {
+                        bool added = discoveredDevices[i].setManufacturerData(key, changedManufacturerData.value(key).variant().toByteArray());
+                        wasNewValue = (wasNewValue || added);
+                    }
+
+                    if (wasNewValue)
+                        updatedFields.setFlag(QBluetoothDeviceInfo::Field::ManufacturerData);
+                }
+
+                if (lowEnergySearchTimeout > 0) {
+                    if (discoveredDevices[i] != info) { // field other than manufacturer or rssi changed
+                        if (discoveredDevices.at(i).name() == info.name()) {
+                            qCDebug(QT_BT_BLUEZ) << "Almost Duplicate " << info.address()
+                                                   << info.name() << "- replacing in place";
+                            discoveredDevices.replace(i, info);
+                            emit q->deviceDiscovered(info);
+                        }
+                    } else {
+                        if (!updatedFields.testFlag(QBluetoothDeviceInfo::Field::None))
+                            emit q->deviceUpdated(discoveredDevices[i], updatedFields);
+                    }
+
+                    return;
+                }
+
+                discoveredDevices.replace(i, info);
+                emit q_ptr->deviceDiscovered(discoveredDevices[i]);
+
+                if (!updatedFields.testFlag(QBluetoothDeviceInfo::Field::None))
+                    emit q->deviceUpdated(discoveredDevices[i], updatedFields);
                 return;
             }
         }
     }
 }
-
 QT_END_NAMESPACE
